#ifndef GENERICPTR_HH
#define GENERICPTR_HH

#include <iterator>
#include <algorithm>
#include <map>
#include <stdlib.h>
#include <utils/startofobject.hh>

namespace utils
{
    template<typename T>
    class GenericPtr;
    template<typename T>
    class GenericWeakPtr;
    
    namespace support
    {
	class GenericPtrHelper
	{
	public:
	    typedef void (*GenericFunctionPtr)();
	    
	    struct Deleter
	    {
		virtual void doDelete(GenericFunctionPtr deleteFunc,
				      void* originalPtr) =0;
	    protected:
		~Deleter(){}
	    };

	    struct CountObject;
	    typedef std::map<void*,CountObject> CountMap;
	    typedef CountMap::iterator CountType;

	    struct CountObject
	    {
		unsigned count;
		Deleter* const deleter;
		GenericFunctionPtr const deleteFunction;
		void* const originalPtr;
		CountType parentCount;
		
		CountObject(unsigned c,
			    Deleter* d,
			    GenericFunctionPtr df,
			    void* op):
		    count(c),
		    deleter(d),
		    deleteFunction(df),
		    originalPtr(op),
		    parentCount(nullCount())
		{}
		    
		CountObject(unsigned c,CountType pc):
		    count(c),
		    deleter(NULL),
		    deleteFunction(NULL),
		    originalPtr(NULL),
		    parentCount(getTop(pc))
		{
		    addRef(parentCount);
		}
		CountObject(const CountObject&other):
		    count(other.count),
		    deleter(other.deleter),
		    deleteFunction(other.deleteFunction),
		    originalPtr(other.originalPtr),
		    parentCount(other.parentCount)
		{
		    addRef(parentCount);
		}
		~CountObject()
		{
		    removeRef(parentCount);
		}

	    private:
		static inline CountType getTop(CountType pc)
		{
		    if(pc==nullCount()
		       || pc->second.parentCount==nullCount())
			return pc;
		    return getTop(pc->second.parentCount);
		}
		CountObject& operator=(const CountObject&);
	    };

	    static CountMap counts;	    

	    static inline CountType nullCount()
	    {
		return counts.end();
	    }
	    static CountType addRef(void* p,
				   Deleter* deleter,
				   GenericFunctionPtr func,
				    void* origPtr);
	    static CountType addRef(void* p,
				    CountType parentCount);
	    static inline void addRef(CountType it)
	    {
		if(it!=nullCount())
		    ++(it->second.count);
	    }
	    static void removeRef(CountType it);
	    
	private:
	    GenericPtrHelper(); // private constructor, not implemented => cannot have instance created
	    GenericPtrHelper(const GenericPtrHelper&); // private copy constructor, not implemented => cannot be copied
	    GenericPtrHelper& operator=(const GenericPtrHelper&); // private assignment operator, not implemented => cannot be assigned
	};
    }
    
    // ref-counted smart pointer to a class which doesn't inherit from Countable
    template<typename T>
    class GenericPtr
    {
	// no harm in making types public (which MSVC requires (at least for ConcreteDeleter)), but I'd rather they were private
    public:
	typedef support::GenericPtrHelper helper;
	typedef helper::CountType CountType;
	typedef void(*DeleteFunctionType)(T*);

	typedef T value_type;
	typedef std::ptrdiff_t difference_type;
	typedef std::random_access_iterator_tag iterator_category;
	typedef T* pointer;
	typedef T& reference;

	struct ConcreteDeleter:
	    public helper::Deleter
	{
	    virtual void doDelete(helper::GenericFunctionPtr deleteFunc,
				  void* originalPtr)
	    {
		DeleteFunctionType df=reinterpret_cast<DeleteFunctionType>(deleteFunc);
		
		T* op=static_cast<T*>(originalPtr);

		if(df)
		    df(op); // call the correctly-cast deletion function with the correctly-cast pointer
	    }
	};

	static void dummyDeleteFunction(T*p)
	{
	    abort();
	}
	static void arrayDeleteFunction(T*p)
	{
	    delete [] p;
	}
	static void defaultDeleteFunction(T*p)
	{
	    delete p;
	}
    private:
	static ConcreteDeleter concreteDeleter;
	
	T*p;
	CountType count;
	GenericPtr(T* rawPtr,CountType parentCount):
	    p(rawPtr),count(helper::addRef(startOfObject(p),parentCount))
	{}
    public:
	typedef T value_type; // T available to the rest of the world
	
	// default constructor creates a new object to count, and counts it
	GenericPtr():
	    p(NULL),count(helper::nullCount())
	{
	}
	// copy constructor
	// only counts if pointer is already counted
	GenericPtr(const GenericPtr&other):
	    p(other.p),count(other.count)
	{
	    helper::addRef(count);
	}

	// constructor that takes a raw pointer 
	explicit GenericPtr(T*rawPtr):
	    p(rawPtr),count(helper::addRef(startOfObject(p),
					   &concreteDeleter,
					   reinterpret_cast<helper::GenericFunctionPtr>(defaultDeleteFunction),
					   rawPtr))
	{
	}
	GenericPtr(T*rawPtr,DeleteFunctionType suppliedDeleteFunction):
	    p(rawPtr),count(helper::addRef(startOfObject(p),
					   &concreteDeleter,
					   reinterpret_cast<helper::GenericFunctionPtr>(suppliedDeleteFunction),
					   rawPtr))
	{
	}
	// when destroying, if we have a counted object, decrease count, and destroy if no counts left
	// Throw here is bad anyway, so empty throw spec ensures terminate
	~GenericPtr() throw()
	{
	    helper::removeRef(count);
	}

	// swap two pointers
	void swap(GenericPtr&other) throw()
	{
	    std::swap(p,other.p);
	    std::swap(count,other.count);
	}
	// assignment
	GenericPtr& operator=(const GenericPtr&other)
	{
	    GenericPtr temp(other);
	    swap(temp);
	    return *this;
	}
	// raw pointer based - permits automatic assignment 
	// from derived class smart pointers to base class smart pointers
	GenericPtr& operator=(T* other)
	{
	    GenericPtr temp(other);
	    swap(temp);
	    return *this;
	}
	// additive assignment operators, for array cases
	GenericPtr& operator+=(difference_type diff)
	{
	    GenericPtr temp(*this+diff);
	    swap(temp);
	    return *this;
	}
	GenericPtr& operator-=(difference_type diff)
	{
	    GenericPtr temp(*this-diff);
	    swap(temp);
	    return *this;
	}
	// additive operators
	GenericPtr operator+(difference_type diff) const
	{
	    return GenericPtr(p+diff,count);
	}
	friend GenericPtr operator+(difference_type diff,const GenericPtr& rhs)
	{
	    return GenericPtr(rhs.p+diff,count);
	}
	GenericPtr operator-(difference_type diff) const
	{
	    return GenericPtr(p-diff,count);
	}
	friend GenericPtr operator-(difference_type diff,const GenericPtr& rhs)
	{
	    return GenericPtr(rhs.p-diff,count);
	}
	friend difference_type operator-(const GenericPtr& lhs,const GenericPtr& rhs)
	{
	    return lhs.p-rhs.p;
	}
	// pre- and post- increment and decrement
	GenericPtr& operator++() // pre
	{
	    GenericPtr temp(*this+1);
	    swap(temp);
	    return *this;
	}
	GenericPtr operator++(int) //post
	{
	    GenericPtr temp(*this+1);
	    swap(temp);
	    return temp;
	}
	GenericPtr& operator--() // pre
	{
	    GenericPtr temp(*this-1);
	    swap(temp);
	    return *this;
	}
	GenericPtr operator--(int) //post
	{
	    GenericPtr temp(*this-1);
	    swap(temp);
	    return temp;
	}
	// dereference
	T& operator*() const
	{
	    return *p;
	}
	T* operator->() const
	{
	    return p;
	}
	T& operator[](int index) const
	{
	    return p[index];
	}

	// convert to raw pointer
	T* get() const
	{
	    return p;
	}

	// check for NULL
	bool operator!() const
	{
	    return !p;
	}

	// set to NULL
	void reset()
	{
	    GenericPtr temp(NULL,NULL);
	    swap(temp);
	}
    };

    template<typename T>
    GenericPtr<T>::ConcreteDeleter GenericPtr<T>::concreteDeleter;
}

// overload std::swap to work with GenericPtr
namespace std
{
    template<typename T>
    void swap(::utils::GenericPtr<T>& lhs,::utils::GenericPtr<T>& rhs) throw()
    {
	lhs.swap(rhs);
    }
}


#endif
